<?php

namespace Drupal\media_folders\Controller;

use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Drupal\Core\Controller\ControllerBase;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Cache\CacheableJsonResponse;
use Drupal\Core\Cache\CacheableMetadata;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\RequestStack;
use Drupal\media_folders\MediaFoldersUiBuilder;
use Drupal\media_folders\MediaFoldersUiActions;

/**
 * Class File media folder file Controller.
 */
class MediaFoldersController extends ControllerBase {

  /**
   * The folders ui service.
   *
   * @var \Drupal\media_folders\MediaFoldersUiBuilder
   */
  protected $foldersUi;

  /**
   * The folders ui actions service.
   *
   * @var \Drupal\media_folders\MediaFoldersUiActions
   */
  protected $foldersActionsUi;

  /**
   * The request stack service.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('media_folders.ui_builder'),
      $container->get('media_folders.ui_actions'),
      $container->get('request_stack')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function __construct(
    MediaFoldersUiBuilder $media_folders_ui,
    MediaFoldersUiActions $media_folders_actions_ui,
    RequestStack $request_stack,
  ) {
    $this->foldersUi = $media_folders_ui;
    $this->foldersActionsUi = $media_folders_actions_ui;
    $this->requestStack = $request_stack;
  }

  /**
   * Has media create permission.
   */
  public function hasMediaCreateAccess($folder, $bundle = NULL, $bool = FALSE) {
    return $this->foldersUi->hasMediaCreateAccess($folder);
  }

  /**
   * Get title.
   */
  public function getTitle(mixed $folder = NULL) {
    return $this->foldersUi->getTitle($folder);
  }

  /**
   * Display media folder.
   */
  public function fileExplorer($folder = NULL) {
    if ($folder && ($folder->bundle() !== 'media_folders_folder')) {
      throw new NotFoundHttpException();
    }

    return $this->foldersUi->buildAdminUi($folder);
  }

  /**
   * Display ajax Media folder.
   */
  public function fileExplorerAjax($folder = NULL) {
    if ($folder && ($folder->bundle() !== 'media_folders_folder')) {
      throw new NotFoundHttpException();
    }

    $build = [];
    $output = $this->foldersUi->buildAjaxResponse($folder, $build);
    $response = new CacheableJsonResponse($output, 200);
    $response->addCacheableDependency(CacheableMetadata::createFromRenderArray($build));
    if ($folder) {
      $response->addCacheableDependency($folder);
    }

    return $response;
  }

  /**
   * Display ajax Media folder.
   */
  public function fileExplorerWidgetAjax($folder = NULL, $entity_type_id = NULL, $bundle = NULL, $widget_id = NULL) {
    if ($folder && ($folder->bundle() !== 'media_folders_folder')) {
      throw new NotFoundHttpException();
    }

    $build = [];
    $request = $this->requestStack->getCurrentRequest()->request->all();
    $output = $this->foldersUi->buildWidgetAjaxResponse($folder, $request, $entity_type_id, $bundle, $widget_id, $build);
    $response = new CacheableJsonResponse($output, 200);
    $response->addCacheableDependency(CacheableMetadata::createFromRenderArray($build));
    if ($folder) {
      $response->addCacheableDependency($folder);
    }

    return $response;
  }

  /**
   * Ajax upload files to folder.
   */
  public function ajaxUploadFile($folder) {
    $request = $this->requestStack->getCurrentRequest()->request->all();
    $response = $this->foldersActionsUi->ajaxUploadFile($folder, $request);

    return new JsonResponse($response);
  }

  /**
   * Ajax upload files to folder.
   */
  public function ajaxMoveInto($folder) {
    $request = $this->requestStack->getCurrentRequest()->request->all();
    $response = $this->foldersActionsUi->ajaxMoveInto($folder, $request);

    return new JsonResponse($response);
  }

}
