<?php

namespace Drupal\media_folders;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Entity\EntityTypeBundleInfo;
use Drupal\Core\File\FileExists;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Language\LanguageInterface;
use Drupal\file\FileRepository;
use Drupal\file\Validation\FileValidator;
use Drupal\Core\Transliteration\PhpTransliteration;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\File\FileSystem;
use Drupal\Core\Utility\Token;

/**
 * Service which builds the file explorer widget.
 */
class MediaFoldersUiActions {

  use StringTranslationTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The config service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The bundle info.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfo
   */
  protected $bundleInfo;

  /**
   * File file repository object.
   *
   * @var \Drupal\file\FileRepository
   */
  protected $fileRepository;

  /**
   * File file validator object.
   *
   * @var \Drupal\file\Validation\FileValidator
   */
  protected $fileValidator;

  /**
   * File transliteration object.
   *
   * @var \Drupal\Core\Transliteration\PhpTransliteration
   */
  protected $transliteration;

  /**
   * The language service.
   *
   * @var \Drupal\Core\Language\LanguageInterface
   */
  protected $languageManager;

  /**
   * File file system object.
   *
   * @var \Drupal\file\FileSystem
   */
  protected $fileSystem;

  /**
   * Token object.
   *
   * @var \Drupal\Core\Utility\Token
   */
  protected $token;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    ConfigFactoryInterface $config_factory,
    EntityTypeBundleInfo $bundle_info,
    FileRepository $file_repository,
    FileValidator $file_validator,
    PhpTransliteration $php_transliteration,
    LanguageManagerInterface $languageManager,
    FileSystem $file_system,
    Token $token,
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->configFactory = $config_factory;
    $this->bundleInfo = $bundle_info;
    $this->fileRepository = $file_repository;
    $this->fileValidator = $file_validator;
    $this->transliteration = $php_transliteration;
    $this->languageManager = $languageManager;
    $this->fileSystem = $file_system;
    $this->token = $token;
  }

  /**
   * Ajax upload files to folder.
   */
  public function ajaxMoveInto($folder, $request) {
    $response = [
      'messages' => [],
    ];

    if (!empty($request['objects'])) {
      $folder = $this->entityTypeManager->getStorage('taxonomy_term')->load($folder);
      $folder_id = ($folder) ? $folder->id() : NULL;
      foreach ($request['objects'] as $object) {
        if (!empty($object['type']) && !empty($object['mid'])) {
          if (!$folder || ($folder && $folder->bundle() == 'media_folders_folder')) {
            if ($object['type'] == 'folder') {
              if ($object['mid'] == $folder_id) {
                $response['messages'][] = [
                  'type' => 'error',
                  'message' => $this->t('Can not move folder inside itself!'),
                ];
              }
              else {
                $move = $this->entityTypeManager->getStorage('taxonomy_term')->load($object['mid']);
                if ($move) {
                  if ($folder) {
                    $children = $this->entityTypeManager->getStorage('taxonomy_term')->getChildren($folder);
                  }
                  else {
                    $children = $this->entityTypeManager->getStorage('taxonomy_term')->loadTree('media_folders_folder', 0, 1, TRUE);
                  }
                  if (!empty($children)) {
                    foreach ($children as $child) {
                      if (strtolower($child->getName()) == strtolower($move->getName())) {
                        $response['messages'][] = [
                          'type' => 'error',
                          'message' => $this->t('There is already one folder with the same name!'),
                        ];
                      }
                    }
                  }
                  $move->parent->target_id = $folder_id;
                  $move->save();
                  $response['messages'][] = [
                    'type' => 'status',
                    'message' => $this->t('Folder @folder moved!', ['@folder' => $move->getName()]),
                  ];
                }
              }
            }
            elseif ($object['type'] == 'file') {
              $move = $this->entityTypeManager->getStorage('media')->load($object['mid']);
              if ($move) {
                $query = $this->entityTypeManager->getStorage('media')->getQuery();
                $query->condition('status', TRUE);
                if ($folder) {
                  $query->condition('field_folders_folder', $folder->id());
                }
                else {
                  $query->notExists('field_folders_folder');
                }
                $ids = $query->accessCheck(FALSE)->execute();
                $files = $this->entityTypeManager->getStorage('media')->loadMultiple($ids);
                if (!empty($files)) {
                  foreach ($files as $file) {
                    if (strtolower($file->getName()) == strtolower($move->getName())) {
                      $response['messages'][] = [
                        'type' => 'error',
                        'message' => $this->t('There is already one file with the same name!'),
                      ];
                    }
                  }
                }
                $move->field_folders_folder->target_id = $folder_id;
                $move->save();
                $response['messages'][] = [
                  'type' => 'status',
                  'message' => $this->t('File @filename moved!', ['@filename' => $move->getName()]),
                ];
              }
            }
          }
        }
      }
    }

    return $response;
  }

  /**
   * Ajax upload files to folder.
   */
  public function ajaxUploadFile($folder, $request) {
    $response = [
      'messages' => [],
    ];

    if (!empty($request['request'])) {
      $folder = $this->entityTypeManager->getStorage('taxonomy_term')->load($folder);
      if (!$folder || ($folder && $folder->bundle() == 'media_folders_folder')) {
        $validators = $this->getFileValidators();
        $filenames = $this->getFolderFileNames($folder);

        foreach ($request['request'] as $post) {
          try {
            $media_file_name = $post['name'];
            $position = strrpos($media_file_name, '.');
            $extension = substr($media_file_name, $position);
            $bundle = $this->findBundleByExtension($extension, $validators);
            if (!$bundle) {
              $response['messages'][] = [
                'type' => 'error',
                'message' => $this->t('File not allowed!'),
              ];
              continue;
            }

            $this->fileNameRename($media_file_name, $filenames);
            $files_data = preg_replace('#^data:application/\w+;base64,#i', '', $post['contents']);
            $files_data = preg_replace('#^data:image/\w+;base64,#i', '', $files_data);
            $files_data = preg_replace('#^data:video/\w+;base64,#i', '', $files_data);
            $files_data = preg_replace('#^data:audio/\w+;base64,#i', '', $files_data);
            $files_data = preg_replace('#^data:text/\w+;base64,#i', '', $files_data);
            $file_data = base64_decode($files_data);
            $file = $this->fileRepository->writeData($file_data, 'temporary://' . $this->sanitizeFilename($media_file_name), FileExists::Rename);
            $file->status->value = 0;
            $file->save();

            $invalid = $this->fileValidator->validate($file, $validators[$bundle]);
            if ($invalid->count() > 0) {
              for ($i = 0; $i < $invalid->count(); $i++) {
                $response['messages'][] = [
                  'type' => 'error',
                  'message' => $invalid->get($i)->getMessage()->render(),
                ];
              }
            }
            else {
              $this->createMedia($file, $folder, $bundle, $media_file_name);
              $response['messages'][] = [
                'type' => 'status',
                'message' => $this->t('File @filename uploaded!', ['@filename' => $post['name']]),
              ];
            }
          }
          catch (\Exception $e) {
            $response['messages'][] = [
              'type' => 'error',
              'message' => $e->getMessage(),
            ];
          }
        }
      }
    }

    return $response;
  }

  /**
   * Ajax upload files to folder.
   */
  public function getFileValidators() {
    $validators = [];
    $bundles = $this->bundleInfo->getBundleInfo('media');
    foreach (array_keys($bundles) as $bundle) {
      $field = MediaFoldersUiBuilder::getFolderEntitiesFileField($bundle);
      if ($field) {
        $field_config = $this->entityTypeManager->getStorage('field_config')->load('media.' . $bundle . '.' . $field);
        $field_settings = $field_config->getSettings();

        if (!empty($field_settings['file_extensions'])) {
          $validators[$bundle]['FileExtension'] = ['extensions' => $field_settings['file_extensions']];
        }
        if (!empty($field_settings['max_filesize'])) {
          $validators[$bundle]['FileSizeLimit'] = ['fileLimit' => $field_settings['max_filesize']];
        }
      }
    }

    return $validators;
  }

  /**
   * Ajax upload files to folder.
   */
  public function getFolderFileNames($folder) {
    $filenames = [];
    $properties = [
      'status' => TRUE,
    ];
    if (!empty($folder)) {
      $properties['field_folders_folder'] = $folder->id();
    }
    $files = $this->entityTypeManager->getStorage('media')->loadByProperties($properties);
    if (!empty($files)) {
      foreach ($files as $file) {
        $filenames[$file->getName()] = $file->getName();
      }
    }

    return $filenames;
  }

  /**
   * Ajax upload files to folder.
   */
  public function findBundleByExtension($ext, $validators) {
    $ext = ltrim($ext, '.');
    foreach ($validators as $bundle => $values) {
      if (!empty($values['FileExtension']['extensions'])) {
        $allowed_extensions = explode(' ', $values['FileExtension']['extensions']);
        if (in_array($ext, $allowed_extensions)) {
          return $bundle;
        }
      }
    }

    return FALSE;
  }

  /**
   * Ajax upload files to folder.
   */
  public function fileNameRename(&$filename, $filenames) {
    if (in_array($filename, $filenames)) {
      $pos = strrpos($filename, '.');
      if ($pos !== FALSE) {
        $name = substr($filename, 0, $pos);
        $ext = substr($filename, $pos);
      }
      else {
        $name = $filename;
        $ext = '';
      }

      $counter = 1;
      do {
        $filename = $name . ' (' . $counter++ . ')' . $ext;
      } while (in_array($filename, $filenames));
    }
  }

  /**
   * Ajax upload files to folder.
   */
  public function createMedia($file, $folder, $bundle, $media_file_name) {
    $field = MediaFoldersUiBuilder::getFolderEntitiesFileField($bundle);
    $field_config = $this->entityTypeManager->getStorage('field_config')->load('media.' . $bundle . '.' . $field);
    $field_settings = $field_config->getSettings();

    $destination = $this->token->replace($field_settings['uri_scheme'] . '://' . $field_settings['file_directory']);
    $this->fileSystem->prepareDirectory($destination, FileSystemInterface::CREATE_DIRECTORY);
    $file->status->value = 1;
    $file->save();
    $this->fileRepository->move($file, $destination, FileExists::Rename);

    $folder_file = $this->entityTypeManager->getStorage('media')->create(
      [
        'bundle' => $bundle,
        'name' => $media_file_name,
        $field => [
          'target_id' => $file->id(),
        ],
        'field_folders_folder' => [
          'target_id' => ($folder) ? $folder->id() : NULL,
        ],
        'status' => 1,
      ]
    );
    $folder_file->save();

    return $folder_file;
  }

  /**
   * Sanitize filename.
   */
  private function sanitizeFilename($filename) {
    $fileSettings = $this->configFactory->get('file.settings');
    $transliterate = $fileSettings->get('filename_sanitization.transliterate');
    $extension = pathinfo($filename, PATHINFO_EXTENSION);
    if ($extension !== '') {
      $extension = '.' . $extension;
      $filename = pathinfo($filename, PATHINFO_FILENAME);
    }

    $alphanumeric = $fileSettings->get('filename_sanitization.replace_non_alphanumeric');
    $replacement = $fileSettings->get('filename_sanitization.replacement_character');
    if ($transliterate) {
      $transliterated_filename = $this->transliteration->transliterate(
        $filename,
        $this->languageManager->getCurrentLanguage(LanguageInterface::TYPE_CONTENT)->getId(),
        $replacement
      );
      if (mb_strlen($transliterated_filename) > 0) {
        $filename = $transliterated_filename;
      }
      else {
        $alphanumeric = TRUE;
      }
    }
    if ($fileSettings->get('filename_sanitization.replace_whitespace')) {
      $filename = preg_replace('/\s/u', $replacement, trim($filename));
    }
    if ($transliterate && $alphanumeric) {
      $filename = preg_replace('/[^0-9A-Za-z_.-]/u', $replacement, $filename);
    }
    if ($fileSettings->get('filename_sanitization.deduplicate_separators')) {
      $filename = preg_replace('/(_)_+|(\.)\.+|(-)-+/u', $replacement, $filename);
      $filename = preg_replace('/(_|\.|\-)[(_|\.|\-)]+/u', $replacement, $filename);
      $filename = preg_replace('/' . preg_quote($replacement, NULL) . '[' . preg_quote($replacement, NULL) . ']*/u', $replacement, $filename);
      $filename = rtrim($filename, $replacement);

      if (!empty($extension)) {
        $filename = rtrim($filename, '.');
      }
    }
    if ($fileSettings->get('filename_sanitization.lowercase')) {
      $filename = mb_strtolower($filename);
    }

    return $filename . $extension;
  }

}
