<?php

namespace Drupal\media_folders;

use Drupal\Component\Serialization\Json;
use Drupal\Component\Utility\SortArray;
use Drupal\Component\Utility\Unicode;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Link;
use Drupal\Core\Render\Markup;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Symfony\Component\HttpFoundation\RequestStack;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\File\FileUrlGenerator;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Routing\RedirectDestinationInterface;
use Drupal\Core\Form\FormBuilderInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Entity\EntityTypeBundleInfo;

/**
 * Service which builds the file explorer widget.
 */
class MediaFoldersUiBuilder {

  use StringTranslationTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The currently active request object.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $request;

  /**
   * Current user account.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * The config service.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The media folder file settings.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $mediaFolderFileSettings;

  /**
   * File url generator object.
   *
   * @var \Drupal\Core\File\FileUrlGenerator
   */
  protected $fileUrlGenerator;

  /**
   * Drupal\Core\Render\RendererInterface definition.
   *
   * @var Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * The redirect destination service.
   *
   * @var \Drupal\Core\Routing\RedirectDestinationInterface
   */
  protected $redirectDestination;

  /**
   * Form builder will be used via Dependency Injection.
   *
   * @var \Drupal\Core\Form\FormBuilderInterface
   */
  protected $formBuilder;

  /**
   * The current route match.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected $routeMatch;

  /**
   * The default cache.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cache;

  /**
   * The bundle info.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfo
   */
  protected $bundleInfo;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    RequestStack $request_stack,
    AccountInterface $current_user,
    ConfigFactoryInterface $config_factory,
    FileUrlGenerator $fileUrlGenerator,
    RendererInterface $renderer,
    RedirectDestinationInterface $redirect_destination,
    FormBuilderInterface $form_builder,
    RouteMatchInterface $route_match,
    CacheBackendInterface $cache,
    EntityTypeBundleInfo $bundle_info,
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->request = $request_stack->getCurrentRequest();
    $this->currentUser = $current_user;
    $this->configFactory = $config_factory;
    $this->mediaFolderFileSettings = $config_factory->get('media_folders.settings');
    $this->fileUrlGenerator = $fileUrlGenerator;
    $this->renderer = $renderer;
    $this->redirectDestination = $redirect_destination;
    $this->formBuilder = $form_builder;
    $this->routeMatch = $route_match;
    $this->cache = $cache;
    $this->bundleInfo = $bundle_info;
  }

  /**
   * {@inheritdoc}
   */
  public static function dialogOptions() {
    return [
      'classes' => [
        'ui-dialog' => 'media-folders-widget-modal',
      ],
      'title' => t('Add or select files'),
      'height' => '75%',
      'width' => '75%',
    ];
  }

  /**
   * Has media create permission.
   */
  public function hasMediaCreateAccess($folder, $bundle = NULL, $bool = FALSE) {
    if ($this->currentUser->hasPermission('administer media')) {
      return ($bool) ? TRUE : AccessResult::allowed()->cachePerPermissions();
    }

    if ($this->currentUser->hasPermission('create media')) {
      return ($bool) ? TRUE : AccessResult::allowed()->cachePerPermissions();
    }

    if ($bundle && $this->currentUser->hasPermission('create ' . $bundle . ' media')) {
      return ($bool) ? TRUE : AccessResult::allowed()->cachePerPermissions();
    }

    if (!$bundle) {
      $bundles = $this->bundleInfo->getBundleInfo('media');
      foreach (array_keys($bundles) as $bundle_id) {
        if ($this->currentUser->hasPermission('create ' . $bundle_id . ' media')) {
          return ($bool) ? TRUE : AccessResult::allowed()->cachePerPermissions();
        }
      }
    }

    return ($bool) ? FALSE : AccessResult::forbidden()->cachePerPermissions();
  }

  /**
   * Has media create permission.
   */
  public function hasMediaEditAccess($media, $bool = FALSE) {
    if ($this->currentUser->hasPermission('administer media')) {
      return ($bool) ? TRUE : AccessResult::allowed()->cachePerPermissions();
    }

    if ($this->currentUser->hasPermission('update any media')) {
      return ($bool) ? TRUE : AccessResult::allowed()->cachePerPermissions();
    }

    if ($media->uid->target_id == $this->currentUser->id() && $this->currentUser->hasPermission('update media')) {
      return ($bool) ? TRUE : AccessResult::allowed()->cachePerPermissions();
    }

    if ($this->currentUser->hasPermission('edit any ' . $media->bundle() . ' media')) {
      return ($bool) ? TRUE : AccessResult::allowed()->cachePerPermissions();
    }

    if ($media->uid->target_id == $this->currentUser->id() && $this->currentUser->hasPermission('edit own ' . $media->bundle() . ' media')) {
      return ($bool) ? TRUE : AccessResult::allowed()->cachePerPermissions();
    }

    return ($bool) ? FALSE : AccessResult::forbidden()->cachePerPermissions();
  }

  /**
   * Has media create permission.
   */
  public function hasMediaDeleteAccess($media, $bool = FALSE) {
    if ($this->currentUser->hasPermission('administer media')) {
      return ($bool) ? TRUE : AccessResult::allowed()->cachePerPermissions();
    }

    if ($this->currentUser->hasPermission('delete any media')) {
      return ($bool) ? TRUE : AccessResult::allowed()->cachePerPermissions();
    }

    if ($media->uid->target_id == $this->currentUser->id() && $this->currentUser->hasPermission('delete media')) {
      return ($bool) ? TRUE : AccessResult::allowed()->cachePerPermissions();
    }

    if ($this->currentUser->hasPermission('delete any ' . $media->bundle() . ' media')) {
      return ($bool) ? TRUE : AccessResult::allowed()->cachePerPermissions();
    }

    if ($media->uid->target_id == $this->currentUser->id() && $this->currentUser->hasPermission('delete own ' . $media->bundle() . ' media')) {
      return ($bool) ? TRUE : AccessResult::allowed()->cachePerPermissions();
    }

    return ($bool) ? FALSE : AccessResult::forbidden()->cachePerPermissions();
  }

  /**
   * Get title.
   */
  public function getTitle(mixed $folder = NULL) {
    if (!$folder) {
      return $this->t('Media Folders');
    }
    $element = ($folder && ($folder->bundle() == 'media_folders_folder')) ? $folder->getName() : '';

    return $this->t('Media Folders | @element', ['@element' => $element]);
  }

  /**
   * {@inheritdoc}
   */
  public function buildAdminUi($folder) {
    $folder_id = ($folder) ? $folder->id() : 0;
    $view_type = $this->getCookie('view-type', $this->mediaFolderFileSettings->get('default_view'));
    $cache_tags = [
      'media_folders_page',
      'taxonomy_term:' . $folder_id,
    ];

    $folders = $this->getFolderFolders($folder_id, $view_type, $cache_tags);
    $folders_ordered = $this->itemsOrder($folders);
    $this->getFolderFiles($folder, $folders_ordered, $cache_tags);
    $this->addTableHeader($folders_ordered);
    $navbar = $this->geAllUserFoldersTree($folder_id);
    $actions = $this->getFolderActions($folder);
    $actions_raw = $this->getRawActions($actions);
    $board_class = (!$this->hasMediaCreateAccess(NULL, NULL, TRUE)) ? 'not-droppable' : '';

    return [
      '#theme' => 'media_folders_page',
      '#title' => $this->t('Media folders'),
      '#up_button' => $this->getUpButton($folder),
      '#breadcrumb' => $this->getBreadcrumb($folder),
      '#actions' => $actions,
      '#actions_raw' => ($actions_raw) ? $actions_raw : NULL,
      '#buttons' => Markup::create($this->getButtons()),
      '#form' => Markup::create($this->getForm()),
      '#folders' => [
        '#theme' => 'media_folders_folders',
        '#folders' => $folders_ordered,
      ],
      '#navbar' => [
        '#theme' => 'media_folders_navbar',
        '#folders' => $navbar,
      ],
      '#class' => $view_type,
      '#board_class' => $board_class,
      '#fid' => $folder_id,
      '#attached' => [
        'library' => 'media_folders/folders',
      ],
      '#cache' => [
        'tags' => $cache_tags,
        'contexts' => ['user.roles'],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildAjaxResponse($folder, &$build) {
    $folder_id = ($folder) ? $folder->id() : 0;
    $view_type = $this->getCookie('view-type', $this->mediaFolderFileSettings->get('default_view'));
    $cache_tags = ['taxonomy_term:' . $folder_id];

    $folders = $this->getFolderFolders($folder_id, $view_type, $cache_tags);
    $folders_ordered = $this->itemsOrder($folders);
    $this->getFolderFiles($folder, $folders_ordered, $cache_tags);
    $this->addTableHeader($folders_ordered);
    $actions = $this->getFolderActions($folder);
    $actions_raw = $this->getRawActions($actions);
    $up = $this->getUpButton($folder);
    $folders = [
      '#theme' => 'media_folders_folders',
      '#folders' => $folders_ordered,
    ];
    $navbar = [
      '#theme' => 'media_folders_navbar',
      '#folders' => $this->geAllUserFoldersTree($folder_id),
    ];
    $board_class = (!$this->hasMediaCreateAccess(NULL, NULL, TRUE)) ? 'not-droppable' : '';

    $build['#cache'] = [
      'tags' => $cache_tags,
      'contexts' => ['user.roles'],
    ];

    return [
      'navbar' => '<div class="navbar-header">' . $this->t('Folders') . '</div>' . $this->renderer->render($navbar),
      'folders' => $this->renderer->render($folders) . '<div class="droppable-message">' . $this->t('Drop your file here!') . '</div>',
      'breadcrumb' => $this->getBreadcrumb($folder),
      'actions' => $this->renderer->render($actions),
      'actions_raw' => ($actions_raw) ? '<span class="close"></span>' . $this->renderer->render($actions_raw) : '',
      'buttons' => $this->getButtons(),
      'up_button' => (!$folder) ? $up : $this->renderer->render($up),
      'title' => $this->getTitle($folder),
      'board_class' => $board_class,
      'fid' => $folder_id,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildWidgetUi($state = NULL) {
    if (!$state) {
      $state = MediaFoldersState::fromRequest($this->request);
    }

    $widget_settings = $state->getWidgetSettings();
    $opener_parameters = $state->getOpenerParameters();
    $folder_id = 0;
    $view_type = $widget_settings['view_type'];
    $cache_tags = ['taxonomy_term:' . $folder_id];

    $folders = $this->getFolderFolders($folder_id, $view_type, $cache_tags, FALSE);
    $folders_ordered = $this->itemsOrder($folders, $widget_settings);
    $this->getWidgetFolderFiles(NULL, $folders_ordered, $opener_parameters, $widget_settings);
    $this->addTableHeader($folders_ordered);
    $navbar = $this->geAllUserFoldersTree($folder_id);
    $board_class = 'not-droppable';

    $explorer = [
      '#theme' => 'media_folders_widget',
      '#up_button' => ($widget_settings['show_navigation']) ? $this->getUpButton(NULL) : NULL,
      '#breadcrumb' => ($widget_settings['show_navigation']) ? $this->getBreadcrumb(NULL) : NULL,
      '#form' => ($widget_settings['show_search']) ? Markup::create($this->getForm()) : NULL,
      '#folders' => [
        '#theme' => 'media_folders_folders',
        '#folders' => $folders_ordered,
      ],
      '#navbar' => [
        '#theme' => 'media_folders_navbar',
        '#folders' => $navbar,
      ],
      '#class' => $view_type,
      '#board_class' => $board_class,
      '#fid' => $folder_id,
      '#cache' => [
        'tags' => [
          'taxonomy_term:' . $folder_id,
        ],
      ],
      '#attached' => [
        'drupalSettings' => [
          'media_folders' => [
            'selection_remaining' => $state->getAvailableSlots(),
            'state_parameters' => $opener_parameters,
            'widget_settings' => $widget_settings,
          ],
        ],
      ],
    ];

    return $explorer;
  }

  /**
   * {@inheritdoc}
   */
  public function buildWidgetAjaxResponse($folder, $request, $entity_type_id, $bundle, $widget_id, $build) {
    $folder_id = ($folder) ? $folder->id() : 0;
    $widget_settings = $request['request'];
    $view_type = $widget_settings['view_type'];
    $opener_parameters = [
      'entity_type_id' => $entity_type_id,
      'bundle' => $bundle,
      'field_name' => $widget_id,
    ];
    $cache_tags = ['taxonomy_term:' . $folder_id];

    $folders = $this->getFolderFolders($folder_id, $view_type, $cache_tags, FALSE);
    $folders_ordered = $this->itemsOrder($folders, $widget_settings);
    $this->getWidgetFolderFiles($folder, $folders_ordered, $opener_parameters, $widget_settings);
    $this->addTableHeader($folders_ordered);
    $folders = [
      '#theme' => 'media_folders_folders',
      '#folders' => $folders_ordered,
    ];
    $navbar = [
      '#theme' => 'media_folders_navbar',
      '#folders' => $this->geAllUserFoldersTree($folder_id),
    ];
    $board_class = 'not-droppable';
    $up = $this->getUpButton($folder);
    if ($folder) {
      $up = $this->renderer->render($up);
    }

    $build['#cache'] = [
      'tags' => $cache_tags,
    ];

    return [
      'navbar' => '<div class="navbar-header">' . $this->t('Folders') . '</div>' . $this->renderer->render($navbar),
      'folders' => $this->renderer->render($folders),
      'up_button' => ($widget_settings['show_navigation']) ? $up : NULL,
      'breadcrumb' => ($widget_settings['show_navigation']) ? $this->getBreadcrumb($folder) : NULL,
      'form' => ($widget_settings['show_search']) ? Markup::create($this->getForm()) : NULL,
      'title' => $this->getTitle($folder),
      'board_class' => $board_class,
      'fid' => $folder_id,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function itemsOrder($items, $widget_settings = NULL) {
    $view_order = ($widget_settings) ? $widget_settings['view_order'] : $this->getCookie('view-order', $this->mediaFolderFileSettings->get('default_order'));

    if ($view_order == 'za' || $view_order == 'az') {
      uasort($items, [SortArray::class, 'sortByTitleProperty']);
    }
    elseif ($view_order == 'date-asc' || $view_order == 'date-desc') {
      uasort($items, [SortArray::class, 'sortByWeightProperty']);
    }

    if ($view_order == 'za' || $view_order == 'date-asc') {
      return array_reverse($items);
    }

    return $items;
  }

  /**
   * Get cookie.
   */
  private function getCookie($name, $default) {
    $cookies = $this->request->cookies;
    if ($cookies->has($name)) {
      return $cookies->get($name);
    }

    return $default;
  }

  /**
   * {@inheritdoc}
   */
  public function treeNestedArray(&$build, $tree, $folder_id) {
    foreach ($tree as $term) {
      $key = strtolower($term->getName()) . '-' . $term->id();
      $link = Link::fromTextAndUrl($term->getName(), Url::fromRoute('media_folders.collection.folder', ['folder' => $term->id()]));

      $folder_class = ['navbar-folder'];
      if (!$this->hasMediaCreateAccess($term, NULL, TRUE)) {
        $folder_class[] = 'not-droppable';
      }

      $build[$key] = [
        'id' => $term->id(),
        'link' => $link,
        'class' => ($folder_id == $term->id()) ? 'active' : '',
        'folder_class' => implode(' ', $folder_class),
        'children' => [],
      ];
      $children = $this->entityTypeManager->getStorage('taxonomy_term')->getChildren($term);
      if (!empty($children)) {
        $this->treeNestedArray($build[$key]['children'], $children, $folder_id);
      }
    }
    ksort($build);
  }

  /**
   * {@inheritdoc}
   */
  public function treeNestedArrayActiveTrail(&$build, $folder_id, $paths = FALSE) {
    if ($paths === FALSE) {
      $paths = $this->treeNestedArrayActiveTrailPaths($build, $folder_id);
    }

    if (!empty($paths)) {
      foreach ($paths as $path) {
        if (!empty($build[$path])) {
          $build[$path]['class'] .= ' active-trail';
          $this->treeNestedArrayActiveTrail($build[$path]['children'], $folder_id, $paths);
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function treeNestedArrayActiveTrailPaths($build, $folder_id) {
    $paths = [];
    $iterator = new \RecursiveIteratorIterator(new \RecursiveArrayIterator($build), \RecursiveIteratorIterator::SELF_FIRST);
    foreach ($iterator as $v) {
      if (is_array($v) && !empty($v['id']) && $v['id'] == $folder_id) {
        for ($p = [], $i = 0, $z = $iterator->getDepth(); $i <= $z; $i++) {
          $p = $iterator->getSubIterator($i)->key();
          if ($p != 'children') {
            $paths[] = $p;
          }
        }
      }
    }

    return $paths;
  }

  /**
   * Get user folders.
   */
  public function geAllUserFoldersTree($folder_id) {
    $folders = $this->entityTypeManager->getStorage('taxonomy_term')->loadTree('media_folders_folder', 0, 1, TRUE);
    $build = [];

    $folder_class = ['navbar-folder'];
    if (!$this->hasMediaCreateAccess(NULL, NULL, TRUE)) {
      $folder_class[] = 'not-droppable';
    }
    $build['root'] = [
      'id' => 0,
      'link' => Link::fromTextAndUrl($this->t('Root'), Url::fromRoute('media_folders.collection')),
      'class' => 'root active-trail',
      'folder_class' => implode(' ', $folder_class),
      'children' => [],
    ];

    $this->treeNestedArray($build['root']['children'], $folders, $folder_id);
    $this->treeNestedArrayActiveTrail($build, $folder_id);

    return $build;
  }

  /**
   * Get Folder Files.
   */
  public function getWidgetFolderFiles($folder, &$folders, $opener_parameters, $widget_settings) {
    $files = [];
    $field_config = $this->configFactory->get('field.field.' . $opener_parameters['entity_type_id'] . '.' . $opener_parameters['bundle'] . '.' . $opener_parameters['field_name']);
    $settings = $field_config->get('settings');
    $target_bundles = array_values($settings['handler_settings']['target_bundles']);
    $view_type = $widget_settings['view_type'];
    $values = $this->getFolderEntities($folder, $target_bundles);
    foreach ($values as $file_entity) {
      $field = static::getFolderEntitiesFileField($file_entity->bundle());
      if (!empty($file_entity->$field->entity)) {
        $file = $file_entity->$field->entity;
        $ext = pathinfo($file->getFilename(), PATHINFO_EXTENSION);
        $type = $this->humanFilesize($file->getSize());

        $files[] = [
          '#theme' => 'media_folders_folder',
          '#title' => $file_entity->getName(),
          '#weight' => $file_entity->$field->entity->changed->value,
          '#desc' => $file_entity->$field->description,
          '#link' => $this->fileUrlGenerator->generateAbsoluteString($file->getFileUri()),
          '#classes' => $this->getFileClasses($file),
          '#wrapper_classes' => 'folder-file',
          '#attr' => Markup::create('download data-ext=".' . $ext . '" data-size="' . $type . '"'),
          '#size' => ($view_type == 'list') ? $type : '',
          '#ftype' => ($view_type == 'list') ? $ext : '',
          '#actions' => '',
          '#fid' => $file_entity->id(),
          '#cache' => [
            'tags' => [
              'media:' . $file_entity->id(),
            ],
          ],
        ];
      }
    }
    $files_ordered = $this->itemsOrder($files, $widget_settings);

    $folders = array_merge($folders, $files_ordered);
  }

  /**
   * Get Folder Files.
   */
  private function getFolderFolders($folder_id, $view_type, &$cache_tags, $show_action = TRUE) {
    $folders = [];
    $tree = $this->entityTypeManager->getStorage('taxonomy_term')->loadTree('media_folders_folder', $folder_id, 1, TRUE);
    foreach ($tree as $term) {
      $classes = [
        'folder-icon',
        'folder-icon-folder-empty',
      ];

      $wrapper_classes = ['folder-folder'];
      if (!$this->hasMediaCreateAccess(NULL, NULL, TRUE)) {
        $wrapper_classes[] = 'not-droppable';
      }
      $file_count = $this->getFileCount($term->id());

      $folders[] = [
        '#theme' => 'media_folders_folder',
        '#title' => $term->getName(),
        '#weight' => $term->changed->value,
        '#desc' => (!empty($term->getDescription())) ? Markup::create(strip_tags($term->getDescription())) : '',
        '#link' => Url::fromRoute('media_folders.collection.folder', ['folder' => $term->id()]),
        '#classes' => implode(' ', $classes),
        '#wrapper_classes' => implode(' ', $wrapper_classes),
        '#attr' => Markup::create('data-count="' . $file_count . '"'),
        '#size' => ($view_type == 'list') ? $file_count . ' ' . $this->t('items') : '',
        '#ftype' => ($view_type == 'list') ? $this->t('folder') : '',
        '#actions' => ($show_action) ? Markup::create($this->getItemActions($term)) : '',
        '#fid' => $term->id(),
        '#cache' => [
          'tags' => [
            'taxonomy_term:' . $term->id(),
          ],
        ],
      ];
      $cache_tags[] = 'taxonomy_term:' . $term->id();
    }

    return $folders;
  }

  /**
   * Get Folder Files.
   */
  private function getFolderFiles($folder, &$folders, &$cache_tags) {
    $view_type = $this->getCookie('view-type', $this->mediaFolderFileSettings->get('default_view'));

    $values = $this->getFolderEntities($folder);
    $files = $this->buildFolderFiles($values, $view_type, $cache_tags);
    $files_ordered = $this->itemsOrder($files);

    $folders = array_merge($folders, $files_ordered);
  }

  /**
   * Get Folder Entities.
   */
  public function buildFolderFiles($values, $view_type, &$cache_tags) {
    $files = [];
    foreach ($values as $file_entity) {
      $field = static::getFolderEntitiesFileField($file_entity->bundle());
      if (!empty($file_entity->$field->entity)) {
        $file = $file_entity->$field->entity;
        $ext = pathinfo($file->getFilename(), PATHINFO_EXTENSION);
        $type = $this->humanFilesize($file->getSize());

        $files[] = [
          '#theme' => 'media_folders_folder',
          '#title' => $file_entity->getName(),
          '#weight' => $file_entity->$field->entity->changed->value,
          '#desc' => $file_entity->$field->description,
          '#link' => $this->fileUrlGenerator->generateAbsoluteString($file->getFileUri()),
          '#classes' => $this->getFileClasses($file),
          '#wrapper_classes' => 'folder-file',
          '#attr' => Markup::create('download data-ext=".' . $ext . '" data-size="' . $type . '"'),
          '#size' => ($view_type == 'list') ? $type : '',
          '#ftype' => ($view_type == 'list') ? $ext : '',
          '#actions' => Markup::create($this->getItemActions(FALSE, $file_entity)),
          '#fid' => $file_entity->id(),
          '#cache' => [
            'tags' => [
              'media:' . $file_entity->id(),
            ],
          ],
        ];
        $cache_tags[] = 'media:' . $file_entity->id();
      }
    }

    return $files;
  }

  /**
   * Get Folder Entities.
   */
  public function getFolderEntities($folder, $target_bundles = NULL) {
    $folder_id = ($folder) ? $folder->id() : 0;
    $cid = is_null($target_bundles) ? 'media_folders:files:' . $folder_id : 'media_folders:files:' . implode('_', $target_bundles) . ':' . $folder_id;
    if ($cache = $this->cache->get($cid)) {
      $response = $cache->data;
    }
    else {
      $query = $this->entityTypeManager->getStorage('media')->getQuery();
      if (!is_null($folder)) {
        $query->condition('field_folders_folder', $folder->id());
      }
      else {
        $query->notExists('field_folders_folder');
      }
      if (!is_null($target_bundles)) {
        $query->condition('bundle', $target_bundles, 'IN');
      }
      $query->condition('status', 1);
      $entity_ids = $query->accessCheck(FALSE)->execute();

      if (!empty($entity_ids)) {
        $response = $this->entityTypeManager->getStorage('media')->loadMultiple($entity_ids);
      }
      else {
        $response = [];
      }

      $this->cache->set($cid, $response, CacheBackendInterface::CACHE_PERMANENT, [
        'taxonomy_term_list:media_folders_folder',
        'media_list',
      ]);
    }

    return $response;
  }

  /**
   * Get Folder Entities file field.
   */
  public static function getFolderEntitiesFileField($bundle) {
    $fieldDefinitions = \Drupal::service('entity_field.manager')->getFieldDefinitions('media', $bundle);
    foreach ($fieldDefinitions as $field_name => $field_config) {
      if (str_starts_with($field_name, 'field_')) {
        $type = $field_config->getFieldStorageDefinition()->getType();
        if ($type == 'image' || $type == 'file') {
          return $field_name;
        }
      }
    }

    return FALSE;
  }

  /**
   * Add table header.
   */
  public function addTableHeader(&$folders) {
    $header = [
      '#theme' => 'media_folders_folder',
      '#title' => $this->t('Name'),
      '#weight' => -1,
      '#desc' => '',
      '#link' => '',
      '#classes' => '',
      '#wrapper_classes' => 'table-header',
      '#attr' => '',
      '#size' => $this->t('Size'),
      '#ftype' => $this->t('Type'),
      '#actions' => '',
      '#fid' => '',
    ];
    array_unshift($folders, $header);
  }

  /**
   * Get Up Button.
   */
  public function getUpButton($parent) {
    if (isset($parent->parent->target_id) && $parent->parent->target_id == 0) {
      $link = Link::fromTextAndUrl('', Url::fromRoute('media_folders.collection'))->toRenderable();
      $link['#attributes']['title'] = $this->t('Parent folder');
      $link['#attributes']['data-label'] = $this->t('Parent folder');

      return $link;
    }
    elseif (isset($parent->parent->target_id)) {
      $link = Link::fromTextAndUrl('', Url::fromRoute('media_folders.collection.folder', ['folder' => $parent->parent->target_id]))->toRenderable();
      $link['#attributes']['title'] = $this->t('Parent folder');
      $link['#attributes']['data-label'] = $this->t('Parent folder');

      return $link;
    }

    return Markup::create('<span></span>');
  }

  /**
   * Get Breadcrumb.
   */
  public function getBreadcrumb($parent) {
    if (isset($parent->parent->target_id)) {
      $parents = [];
      $term = clone $parent;
      $parents[$parent->id()] = '<span>' . Unicode::truncate($parent->getName(), 15, FALSE, TRUE) . '</span>';
      while (!empty($term->parent->target_id)) {
        $term = $term->parent->entity;
        $parents[$term->id()] = Link::fromTextAndUrl(Unicode::truncate($term->getName(), 15, FALSE, TRUE), Url::fromRoute('media_folders.collection.folder', ['folder' => $term->id()]))->toString();
      }
      $parents[0] = Link::fromTextAndUrl($this->t('Root'), Url::fromRoute('media_folders.collection'))->toString();
      $parents = array_reverse($parents);

      return Markup::create(implode('<i></i>', $parents));
    }

    return Markup::create('<span class="root">' . $this->t('Root') . '</span>');
  }

  /**
   * Get File Count.
   */
  public function getFileCount($folder) {
    $sub_folder_count = $this->entityTypeManager->getStorage('taxonomy_term')->loadChildren($folder);
    $query = $this->entityTypeManager->getStorage('media')->getQuery();
    $query->condition('field_folders_folder', $folder);
    $query->condition('status', 1);
    $entity_ids = $query->accessCheck(FALSE)->execute();

    return count($sub_folder_count) + count($entity_ids);
  }

  /**
   * Get Item Actions.
   */
  private function getItemActions(mixed $folder = FALSE, mixed $file = FALSE) {
    $items = $render_items = [];

    if ($folder) {
      if ($this->currentUser->hasPermission('edit terms in media_folders_folder')) {
        $items['edit'] = $this->createDialogLink($this->t('Edit folder'), Url::fromRoute('media_folders.edit_folder', ['folder' => $folder->id()]), ['edit-action'], $this->t('Edit'), TRUE, '90%');
      }

      if ($this->currentUser->hasPermission('delete terms in media_folders_folder')) {
        $items['delete'] = $this->createDialogLink($this->t('Delete folder'), Url::fromRoute('media_folders.delete_folder', ['folder' => $folder->id()]), ['delete-action'], $this->t('Delete'), FALSE, '90%');
      }
    }
    elseif ($file) {
      if ($this->hasMediaEditAccess($file, TRUE)) {
        $items['edit'] = $this->createDialogLink($this->t('Edit media'), Url::fromRoute('entity.media.edit_form', ['media' => $file->id()]), ['edit-action'], $this->t('Edit'), TRUE);
      }
      if ($this->currentUser->hasPermission('view all media revisions')) {
        $items['revision'] = $this->createDialogLink($this->t('Media revisions'), Url::fromRoute('entity.media.version_history', ['media' => $file->id()]), ['revisions-action'], $this->t('RevisionsEdit'), TRUE, '70%');
      }
      if ($this->hasMediaDeleteAccess($file, TRUE)) {
        $items['delete'] = $this->createDialogLink($this->t('Delete media'), Url::fromRoute('entity.media.delete_form', ['media' => $file->id()]), ['delete-action'], $this->t('Delete'), TRUE);
      }
    }

    foreach ($items as $item) {
      $render_items[] = $this->renderer->render($item);
    }

    return implode('', $render_items);
  }

  /**
   * Create Dialog Link.
   */
  private function createDialogLink($text, $url, $classes = [], $title = FALSE, $query = FALSE, $width = '40%') {
    $link = $this->createLink($text, $url, $classes, $title, $query);
    $link['#attributes']['class'][] = 'use-ajax';
    $link['#attributes']['data-dialog-type'] = 'modal';
    $link['#attributes']['data-dialog-options'] = Json::encode(['width' => $width]);
    $link['#attached'] = [
      'library' => [
        'core/drupal.dialog.ajax',
      ],
    ];
    return $link;
  }

  /**
   * Create Link.
   */
  private function createLink($text, $url, $classes = [], $title = FALSE, $query = FALSE) {
    $link = Link::fromTextAndUrl($text, $url)->toRenderable();
    $link['#attributes'] = [
      'class' => $classes,
      'title' => ($title ? $title : $text),
    ];
    if ($query) {
      $destination = $this->redirectDestination->getAsArray();
      if (!empty($destination['destination'])) {
        $url = parse_url($destination['destination']);
        $destination['destination'] = str_replace('/ajax', '', $url['path']);
      }
      $link['#url']->setOption('query', $destination);
    }

    return $link;
  }

  /**
   * Get folder actions.
   */
  private function getFolderActions($folder) {
    $action_links = [];

    $destination = $this->redirectDestination->getAsArray();
    if (!empty($destination['destination'])) {
      $url = parse_url($destination['destination']);
      $destination['destination'] = str_replace('/ajax', '', $url['path']);
    }

    if (!$folder && $this->currentUser->hasPermission('create terms in media_folders_folder')) {
      $action_links['new'] = [
        'title' => $this->t('New folder'),
        'url' => Url::fromRoute('media_folders.add_root_folder', [], ['query' => $destination]),
        'class' => ['new-folder'],
      ];
      $this->transformDialogLink($action_links['new'], $this->t('New folder'), TRUE, '90%');
    }
    elseif ($folder && $this->currentUser->hasPermission('create terms in media_folders_folder')) {
      $action_links['new'] = [
        'title' => $this->t('New folder'),
        'url' => Url::fromRoute('media_folders.add_folder', ['folder' => $folder->id()], ['query' => $destination]),
        'class' => ['new-folder'],
      ];
      $this->transformDialogLink($action_links['new'], $this->t('New folder'), TRUE, '90%');
    }

    if ($this->hasMediaCreateAccess(NULL, NULL, TRUE)) {
      $action_links['upload'] = [
        'title' => $this->t('Upload files'),
        'url' => Url::fromRoute('media_folders.add_file', [
          'folder' => ($folder) ? $folder->id() : 0,
        ], ['query' => $destination]),
        'class' => ['upload-file  '],
      ];
      $this->transformDialogLink($action_links['upload'], $this->t('Upload files'), TRUE, '70%');
    }

    if ($folder && $this->currentUser->hasPermission('edit terms in media_folders_folder')) {
      $action_links['edit'] = [
        'title' => $this->t('Edit folder'),
        'url' => Url::fromRoute('media_folders.edit_folder', ['folder' => $folder->id()], ['query' => $destination]),
        'class' => ['edit-action'],
      ];
      $this->transformDialogLink($action_links['edit'], $this->t('Edit folder'), TRUE, '70%');
    }

    if ($folder && $this->currentUser->hasPermission('delete terms in media_folders_folder')) {
      $action_links['delete'] = [
        'title' => $this->t('Delete folder'),
        'url' => Url::fromRoute('media_folders.delete_folder', ['folder' => $folder->id()]),
        'class' => ['delete-action'],
      ];
      $this->transformDialogLink($action_links['delete'], $this->t('Delete folder'), TRUE, '70%');
    }

    if ($this->currentUser->hasPermission('access taxonomy overview')) {
      $action_links['manage'] = [
        'title' => $this->t('Manage folders'),
        'url' => Url::fromRoute('entity.taxonomy_vocabulary.overview_form', ['taxonomy_vocabulary' => 'media_folders_folder'], ['query' => $destination]),
        'class' => ['manage-folder'],
      ];
      $this->transformDialogLink($action_links['manage'], $this->t('Manage folders'), TRUE);
    }

    if (!empty($action_links)) {
      $actions = [
        '#type' => 'dropbutton',
        '#links' => $action_links,
      ];

      return $actions;
    }

    return '';
  }

  /**
   * Transform Dialog Link.
   */
  private function transformDialogLink(&$link, $title = FALSE, $query = FALSE, $width = '40%') {
    $link['attributes']['class'][] = 'use-ajax';
    $link['attributes']['data-dialog-type'] = 'modal';
    $link['attributes']['data-dialog-options'] = Json::encode(['width' => $width]);
    $link['#attached'] = [
      'library' => [
        'core/drupal.dialog.ajax',
      ],
    ];
  }

  /**
   * Get Raw actions.
   */
  public function getRawActions($actions) {
    $actions_raw = [];
    if (!empty($actions['#links'])) {
      foreach ($actions['#links'] as $key => $link) {
        $actions_raw[$key] = [
          '#type'  => 'link',
          '#title' => $link['title'],
          '#url'  => $link['url'],
          '#attributes' => $link['attributes'],
          '#attached' => $link['#attached'],
        ];
        $actions_raw[$key]['#attributes']['class'][] = $link['class'][0];
      }
    }

    return $actions_raw;
  }

  /**
   * Get Buttons.
   */
  private function getButtons() {
    $action_links = [];

    $destination = $this->redirectDestination->getAsArray();
    if (!empty($destination['destination'])) {
      $url = parse_url($destination['destination']);
      $destination['destination'] = str_replace('/ajax', '', $url['path']);
    }

    $action_links['sorts'] = [
      'title' => $this->t('Sort by'),
      'url' => Url::fromUri('internal://<nolink>'),
      'attributes' => [
        'class' => ['sorts'],
      ],
    ];
    $action_links['az'] = [
      'title' => $this->t('Alphabetic ascending'),
      'url' => Url::fromRoute('media_folders.cookies', [
        'name' => 'view-order',
        'value' => 'az',
      ], ['query' => $destination]),
      'attributes' => [
        'class' => ['az'],
      ],
    ];
    $action_links['za'] = [
      'title' => $this->t('Alphabetic descending'),
      'url' => Url::fromRoute('media_folders.cookies', [
        'name' => 'view-order',
        'value' => 'za',
      ], ['query' => $destination]),
      'attributes' => [
        'class' => ['za'],
      ],
    ];
    $action_links['date-asc'] = [
      'title' => $this->t('Date ascending'),
      'url' => Url::fromRoute('media_folders.cookies', [
        'name' => 'view-order',
        'value' => 'date-asc',
      ], ['query' => $destination]),
      'attributes' => [
        'class' => ['date-asc'],
      ],
    ];
    $action_links['date-desc'] = [
      'title' => $this->t('Date descending'),
      'url' => Url::fromRoute('media_folders.cookies', [
        'name' => 'view-order',
        'value' => 'date-desc',
      ], ['query' => $destination]),
      'attributes' => [
        'class' => ['date-desc'],
      ],
    ];

    $order = $this->getCookie('view-order', $this->mediaFolderFileSettings->get('default_order'));
    if (!empty($order)) {
      $action_links[$order]['attributes']['class'][] = 'active';
      $action_links['sorts']['attributes']['class'][] = $order;
    }

    $actions = [
      '#type' => 'dropbutton',
      '#links' => $action_links,
    ];

    $buttons = $render_buttons = [];
    $buttons['thumbs'] = $this->createLink($this->t('Grid'), Url::fromRoute('media_folders.cookies', [
      'name' => 'view-type',
      'value' => 'thumbs',
    ]), ['thumbs'], $this->t('Folder view'), TRUE);
    $buttons['list'] = $this->createLink($this->t('List'), Url::fromRoute('media_folders.cookies', [
      'name' => 'view-type',
      'value' => 'list',
    ]), ['list'], $this->t('List view'), TRUE);
    $view = $this->getCookie('view-type', $this->mediaFolderFileSettings->get('default_view'));
    $buttons[$view]['#attributes']['class'][] = 'active';

    foreach ($buttons as $button) {
      $render_buttons[] = $this->renderer->render($button);
    }

    return $this->renderer->render($actions) . implode('', $render_buttons);
  }

  /**
   * Get Form.
   */
  public function getForm() {
    $form = $this->formBuilder->getForm('\Drupal\media_folders\Form\MediaFolderFileSearchForm');

    return $this->renderer->render($form);
  }

  /**
   * Get human Filesize.
   */
  public function humanFilesize($bytes, $decimals = 2) {
    $size = ['B', 'kB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'];
    $factor = floor((strlen($bytes) - 1) / 3);

    return (!empty($size[$factor])) ? sprintf("%.{$decimals}f", $bytes / pow(1024, $factor)) . $size[$factor] : '';
  }

  /**
   * {@inheritdoc}
   */
  public function getFileClasses($file) {
    $classes = [
      'folder-icon',
      'folder-icon-file',
    ];

    switch ($file->getMimeType()) {
      case 'text/plain':
        $classes[] = 'folder-icon-file-code';
        break;

      case 'application/pdf':
        $classes[] = 'folder-icon-file-pdf';
        break;

      case 'application/msword':
      case 'application/vnd.openxmlformats-officedocument.wordprocessingml.document':
      case 'application/vnd.oasis.opendocument.text':
        $classes[] = 'folder-icon-file-word';
        break;

      case 'application/vnd.oasis.opendocument.spreadsheet':
      case 'application/vnd.ms-excel':
      case 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet':
        $classes[] = 'folder-icon-file-excel';
        break;

      case 'application/vnd.ms-powerpoint':
      case 'application/vnd.openxmlformats-officedocument.presentationml.presentation':
      case 'application/vnd.oasis.opendocument.presentation':
        $classes[] = 'folder-icon-file-powerpoint';
        break;

      case 'application/gzip':
      case 'application/x-tar':
      case 'application/rar':
      case 'application/zip':
      case 'application/octet-stream':
      case 'application/x-zip-compressed':
      case 'multipart/x-zip':
      case 'application/x-rar-compressed':
      case 'application/vnd.rar':
      case 'application/zip':
        $classes[] = 'folder-icon-file-zip';
        break;

      case 'image/webp':
      case 'image/png':
      case 'image/gif':
      case 'image/jpeg':
      case 'image/jpg':
      case 'image/bmp':
        $classes[] = 'folder-icon-file-image';
        break;

      case 'video/mp4':
      case 'video/webm':
        $classes[] = 'folder-icon-file-video';
        break;

      case 'audio/mpeg':
      case 'audio/x-wav':
        $classes[] = 'folder-icon-file-audio';
        break;

      default:
        $classes[] = 'folder-icon-doc';
        break;
    }

    return implode(' ', $classes);
  }

}
